OBJ
  eeprom : "Basic_I2C_Driver"
  ser : "cserial"

CON
' file system constants
  magic_word   = $45454653
  disk_start   = $0040
  disk_end     = $8000
  block_size   = 64
  eeprom_pin   = 28
  device_sel   = $a0
  attrib_mask  = $3f
  index_mask   = $ffffffc0
  dirent_size  = 30
  fname_size   = 22

' byte offsets for handle variables
  s_index      =  0
  s_offset     =  4
  s_position   =  8
  s_filesize   = 12
  s_cached     = 16
  s_bufptr     = 20
  s_dirtylow   = 24
  s_dirtyhigh  = 28
  s_dirindex   = 32
  s_diroffset  = 36

  handle_size  = 40
  handle_longs = 10

DAT
  first_free long disk_start + block_size
  directory_block long 0[handle_longs]
  directory_buf long 0[block_size/4]
  currentwd_block long 0[handle_longs]
  currentwd_buf long 0[block_size/4]
  readdir_block long 0[8] 'attrib, size, fname
  currentwd_path byte 0[100]

'****************************************************************
' Public API Methods
'****************************************************************
PUB handlesize
{{
  This routine returns the size of a file handle in bytes so that
  the calling routine can allocate a handle that is used by most
  of the file access routines.
}}
  return handle_size + block_size

PUB mount | index
{{
  This routine initializes the disk for file access.  It returns
  a one if the disk is ready, or a zero if it is not.
}}
  eeprom.initialize(eeprom_pin)
  if (ReadEEPromLong(disk_start) == magic_word)
    result := 1
  InitRootHandle(@currentwd_block)
  bytemove(@currentwd_path, string("/"), 2)

PUB format | index
{{
  This routine formats the disk with an empty file system.
}}
  WriteEEPromLong(disk_start, magic_word)
  index := disk_start + block_size
  WriteEEPromLong(disk_start + 4, index | 1)
  WriteEEPromLong(disk_start + 8, 0)
  WriteEEPromLong(index, 0)
  index += block_size
  first_free := index
  repeat while (index < disk_end)
    WriteEEPromLong(index, 1)
    index += block_size
  InitRootHandle(@currentwd_block)
  
PUB remove(fname) | index, size, next1, dirhand2[handle_longs]
{{
  This routine removes a file from the files system.  It returns
  a value of zero if the file did not exist, or a one if it did.
}}
  'ser.str(string("eefs.remove "))
  'ser.str(fname)
  'ser.tx(13)
  index := OpenRead(fname, @dirhand2)
  ifnot index
    return
  ' Remove file from the directory by clearing it's file index
  index &= index_mask
  WriteDirectoryValue(@dirhand2, index, 0)
  ' Free up all the blocks
  repeat while (index)
    next1 := ReadEEPromLong(index)
    WriteEEPromLong(index, next1 | 1)
    if index < first_free
      first_free := index
    index := next1
  return 1
  
PUB rename(fname1, fname2) | filesize
{{
  This routine renames a file in the files system.  It returns
  a value of zero if the file did not exist, or a one if it did.
}}
  ifnot FindFile(fname1, @currentwd_block, @filesize)
    return
  long[@currentwd_block + s_position] -= fname_size
  long[@currentwd_block + s_offset] -= fname_size
  write(@currentwd_block, fname2, fname_size)
  flush(@currentwd_block)
  return 1

PUB open(handle, fname, mode) | index
{{
  This routine opens a file for access.  The access mode
  is determined by the first character of the string "mode".
  The allowable values for the first character are "r", "w"
  and "a" for read, write and append respectively.  The
  pass value of the handle pointer is return if successful.
}}
  'ser.dbprintf2(string("open %s %s"), fname, mode)
  index := 0
  if (handle == 0)
    return
  if (byte[mode] == "r")
    index := OpenRead(fname, handle)
  elseif (byte[mode] == "w")
    index := OpenWrite(fname, handle)
  elseif (byte[mode] == "a")
    index := OpenAppend(fname, handle)
  'ser.dbprintf1(string("retval = %d"), index)
  if (index == 0)
    return
  return handle

PUB close(handle)
  flush(handle)
  'Write File Size

' Thie routine reads a chunk of bytes from an EEProm file
PUB read(handle, buffer, num) | i, position, filesize, value, index, offset, bufptr, cached, num1, watch
{{
  This routine reads "num" bytes from the file defined by "handle".
  The bytes are copies into the memory pointed to by "buffer".
  The return value consist of the number of bytes that were actually
  read.  A zero return value indicates an end of file or an error
  condition.
}}
  'watch := handle + handle_size + block_size
  'ser.dbprintf1(string("read %d"), num)
  'ser.dbprintf2(string("DEBUG1: long[$%x] = $%x"), watch, long[watch])
  if (handle == 0)
    return 0
  if (long[handle + s_index] == 0)
    return 0
  position := long[handle + s_position]
  filesize := long[handle + s_filesize]
  index := long[handle + s_index]
  offset := long[handle + s_offset]
  cached := long[handle + s_cached]
  bufptr := long[handle + s_bufptr]
  'ser.dbprintf4(string("pos = %d, size = %d, index = $%x, offset = %d"), position, filesize, index, offset)
  i := 0
  repeat while (i < num AND position < filesize)
    if (offset => block_size)
      if cached <> index
        index := ReadEEPromLong(index)
      else
        index := long[bufptr]
      offset := 4
    if cached <> index
      'ser.dbprintf2(string("ReadPage: $%x, $%x"), index, bufptr)
      eeprom.ReadPage(eeprom_pin, device_sel, index, bufptr, block_size)
      cached := index
    num1 := num - i
    if (num1 > filesize - position)
      num1 := filesize - position
    if (num1 > block_size - offset)
      num1 := block_size - offset
    'ser.dbprintf3(string("bytemove($%x, $%x, %d)"), buffer+i, bufptr+offset, num1)
    bytemove(buffer + i, bufptr + offset, num1)
    position += num1
    offset += num1
    i += num1
  'ser.dbprintf2(string("DEBUG2: long[$%x] = $%x"), watch, long[watch])
  long[handle + s_position] := position
  long[handle + s_offset] := offset
  long[handle + s_index] := index
  long[handle + s_cached] := cached
  return i

PRI WriteDirtyBlock(index, buffer, pdirtylow, pdirtyhigh) | dirtylow, dirtyhigh
  dirtylow := long[pdirtylow]
  dirtyhigh := long[pdirtyhigh]
  'ser.dbprintf3(string("WriteDirtyBlock: index = $%x, dirtylow = %d, dirtyhigh = %d"), index, dirtylow, dirtyhigh)
  if dirtylow < 0 or dirtyhigh > block_size
    ser.dbprintf2(string("WriteDirtyBlock: ERROR - dirtylow = %d, dirtyhigh = %d"), dirtylow, dirtyhigh)
    return
  if dirtylow => dirtyhigh
    return
  MemCpyEEProm(index + dirtylow, buffer + dirtylow, dirtyhigh - dirtylow)
  long[pdirtylow] := block_size
  long[pdirtyhigh] := 0

PUB write(handle, buffer, num) | value, index, offset, position, filesize, cached, bufptr, dirtylow, dirtyhigh, space, len1, next1, len
{{
  This routine writes "num" bytes to the file defined by "handle".
  The bytes are copied from the memory pointed to by "buffer".
  The return value consist of the number of bytes that were actually
  written.  A zero return value indicates an error occurred.
}}
  'ser.dbprintf1(string("write %d"), num)
  if (handle == 0)
    return -1
  if (num < 1)
    return 0
  len := num
  position := long[handle + s_position]
  filesize := long[handle + s_filesize]
  index := long[handle + s_index]
  offset := long[handle + s_offset]
  cached := long[handle + s_cached]
  bufptr := long[handle + s_bufptr]
  dirtylow := long[handle + s_dirtylow]
  dirtyhigh := long[handle + s_dirtyhigh]
  if (index == 0)
    index := AllocateBlock
    WriteEEPromLong(index, 0)
    'ser.dbprintf1(string("write0: allocated index = $%x"), index)
    WriteDirectoryValue(handle, index | %0110, 0)
    cached := index
    long[bufptr] := 0
    dirtylow := block_size
    dirtyhigh := 0
  'ser.dbprintf4(string("pos = %d, size = %d, index = $%x, offset = %d"), position, filesize, index, offset)
  repeat
    space := block_size - offset
    len1 := len
    if (len1 > space)
      len1 := space
    'MemCpyEEProm(index + offset, buffer, len1)
    bytemove(bufptr + offset, buffer, len1)
    if offset < dirtylow
      dirtylow := offset
    len -= len1
    position += len1
    offset += len1
    if offset > dirtyhigh
      dirtyhigh := offset
    'filesize += len1
    if (len =< 0)
      quit
    buffer += len1
    next1 := long[bufptr]
    if next1
      'ser.dbprintf1(string("write1: existing index = $%x"), next1)
      WriteDirtyBlock(index, bufptr, @dirtylow, @dirtyhigh)
      eeprom.ReadPage(eeprom_pin, device_sel, next1, bufptr, block_size)
    else
      next1 := AllocateBlock
      'ser.dbprintf1(string("write1: allocated index = $%x"), next1)
      'ser.dbprintf1(string("Allocate block $%x"), next1)
      if (next1 == 0)
        num := 0
        quit
     ' WriteEEPromLong(index, next1)
      long[bufptr] := next1
      dirtylow := 0
      if 4 > dirtyhigh
        dirtyhigh := 4
      WriteDirtyBlock(index, bufptr, @dirtylow, @dirtyhigh)
      WriteEEPromLong(next1, 0)
      'dirtyhigh := 0
      'dirtylow := block_size
      long[bufptr] := 0
    index := next1
    cached := index
    offset := 4
  WriteDirtyBlock(index, bufptr, @dirtylow, @dirtyhigh)
  if position > filesize
    filesize := position
    WriteDirectoryValue(handle, filesize, 4)
  long[handle + s_position] := position
  long[handle + s_offset] := offset
  long[handle + s_index] := index
  long[handle + s_filesize] := filesize
  long[handle + s_dirtylow] := dirtylow
  long[handle + s_dirtyhigh] := dirtyhigh
  return num

PUB opendir(path)
{{
  This routine opens the current directory for reading by readdir.
  A handle is returned, which is used by subsequent call by readdir.
}}
  'PrintHandle(@currentwd_block)
  longmove(@directory_block, @currentwd_block, handle_longs)
  'InitRootHandle(@directory_block)
  long[@directory_block + s_bufptr] := @directory_buf
  seek(@directory_block, 0)
  ClearCached(@directory_block)
  'PrintHandle(@directory_block)
  'long[pavail] := DISK_END - disk_start - block_size
  return @directory_block

PUB readdir(dirhand) | index, filesize, filename
{{
  This routine reads the information for the next file in the
  directory indicated by the handle "dirhand".  It returns a
  pointer to the file directory information block if the end of
  the directory has not been reached.  Otherwise it will return
  a zero.  The file directory information block contains 3 longs
  with the filesize, attributes, a physical size followed by the
  file name.
}}
  if dirhand == 0
    return
  repeat
    if read(dirhand, @index, 4) <> 4
      return
    if read(dirhand, @filesize, 4) <> 4
      return
    {
    filename := @readdir_block[3]
    repeat 40
      if read(dirhand, filename, 1) <> 1
        return
      ifnot byte[filename]
        quit
      filename++
    byte[filename] := 0
    }
    if read(dirhand, @readdir_block[2], fname_size) <> fname_size
      return
    'quit 'Uncomment this line to display deleted files
    if index & attrib_mask
      quit

  ifnot index & attrib_mask
    readdir_block[0] := 0
    readdir_block[1] := 0
  else
    readdir_block[0] := index
    readdir_block[1] := filesize
  return @readdir_block

PUB closedir(dirhand)
{{
  This routines closes the directory that had previously been
  opened by opendir.
}}

PUB chdir(path) | i, dirindex, diroffset, index, filesize
{{
  This routine changes the current working directory to the path
  indicated by "dlist".  "dlist" is an array of string pointers
  with "num" entries, which indicate the directory path.  The
  root directory is selected if num is equal to zero.  A value
  of one is returned on success, and zero is returned if the
  path does not exist.
}}
  'ser.dbprintf1(string("num = %d"), num)
  {
  repeat i from 0 to num - 1
    ser.dbprintf1(string("%s"), long[dlist][i])
  }
  if byte[path] == 0 or strcomp(path, string("/"))
    InitRootHandle(@currentwd_block)
    bytemove(@currentwd_path, string("/"), 2)
    return
  index := FindFile(path, @currentwd_block, @filesize)
  ifnot index
    return
  dirindex := long[@currentwd_block + s_index]
  diroffset := long[@currentwd_block + s_offset] - dirent_size
  InitFileHandle(@currentwd_block, dirindex, diroffset, index, filesize, @currentwd_buf)
  return 1

PUB getfilesize(handle)
{{
  This routine returns the size of the file in bytes.
}}
  ifnot handle
    return
  result := long[handle + s_filesize]

PUB seek(handle, newpos) | position, filesize, index, offset, num
{{
  This routine moves to the location specified by "newpos".
}}
  ifnot handle
    return

  position := long[handle + s_position]
  filesize := long[handle + s_filesize]
  index := long[handle + s_index]
  offset := long[handle + s_offset]

  if (newpos < 0) or (newpos > filesize)
    return 0

  ' Backup to the beginning of the current block
  position -= offset - 4
  
  ' Backup to the beginning of the file if newpos < position
  if newpos < position
    index := ReadDirectoryValue(handle, 0) & index_mask
    position := 0

  ' Advance to the new position
  num := newpos - position
  repeat while num > block_size - 4
    index := ReadEEPromLong(index)
    num -= block_size - 4

  long[handle + s_position] := newpos
  long[handle + s_offset] := num + 4
  long[handle + s_index] := index
  return 1

PUB tell(handle)
{{
  This routine returns the current position from the beginning of the file.
}}
  ifnot handle
    return
  result := long[handle + s_position]

PUB flush(handle) | index, cached, bufptr, dirtylow, dirtyhigh
{{
  This routine writes out the data that is sitting in the write buffer.
}}
  ifnot handle
    return
  index := long[handle + s_index]
  cached := long[handle + s_cached]
  bufptr := long[handle + s_bufptr]
  dirtylow := long[handle + s_dirtylow]
  dirtyhigh := long[handle + s_dirtyhigh]
  WriteDirtyBlock(index, bufptr, @dirtylow, @dirtyhigh)
  long[handle + s_dirtylow] := dirtylow
  long[handle + s_dirtyhigh] := dirtyhigh
  'ser.dbprintf2(string("flush: dirtylow = %d, dirtyhight = %d"), long[handle + s_dirtylow], long[handle + s_dirtyhigh])

PUB mkdir(fname)
  return CreateFile(fname, %1111)

PUB chmod(fname, modebits) | index, filesize
  ifnot (index := FindFile(fname, @currentwd_block, @filesize))
    return
  long[@currentwd_block + s_position] -= dirent_size
  long[@currentwd_block + s_offset] -= dirent_size
  index := (index & index_mask) | %10000 | (modebits & %1111)
  write(@currentwd_block, @index, 4)
  flush(@currentwd_block)

'****************************************************************
' Private Methods
'****************************************************************
PRI ClearCached(handle)
  long[handle + s_cached]    := -1
  long[handle + s_dirtylow]  := block_size
  long[handle + s_dirtyhigh] := 0
  
PRI RewindFile(handle)
  long[handle + s_index] := ReadDirectoryValue(handle, 0) & index_mask
  long[handle + s_offset] := 4
  long[handle + s_position] := 0

{
PRI PrintHandle(handle)
  ser.dbprintf5(string("PrintHandle: %x %d %x %d %d"), long[handle + s_dirindex], {
  } long[handle + s_diroffset], long[handle + s_index], long[handle + s_offset], long[handle + s_filesize])
  ser.dbprintf5(string("PrintHandle: %d %x %x %d %d"), long[handle + s_position], {
  } long[handle + s_cached], long[handle + s_bufptr], long[handle + s_dirtylow], long[handle + s_dirtyhigh])
}

PRI InitFileHandle(handle, dirindex, diroffset, index, size, bufptr)
{{
  This routine intializes a file handle with the directory location
  and file information for a file.  The handle is initialized for
  the beginning of the file.
}}
  'ser.dbprintf4(string("InitFileHandle: $%x %d $%x %d"), dirindex, diroffset, index, size)
  long[handle + s_dirindex]  := dirindex
  long[handle + s_diroffset] := diroffset
  long[handle + s_index]     := index & index_mask
  long[handle + s_offset]    := 4
  long[handle + s_filesize]  := size
  long[handle + s_position]  := 0
  long[handle + s_cached]    := -1
  long[handle + s_bufptr]    := bufptr
  long[handle + s_dirtylow]  := block_size
  long[handle + s_dirtyhigh] := 0
  return handle

PRI InitRootHandle(handle) | size
{{
  This routine initializes a file handle for to point to the beginning
  of the root directory.
}}
  size := ReadEEPromLong(disk_start + 8)  ' Get root directory size 
  InitFileHandle(handle, disk_start, 4, disk_start+block_size, size, @currentwd_buf)
  return handle

PRI WriteDirectoryValue(handle, value, offset) | index
{{
  This routine overwrites a long value in the directory entry for a
  file.  The byte offset from the beginning of the directory entry is
  indicated by the value of "offset".  This would have a value of
  zero for the file's starting block address and four for the file
  size.
}}
  index := long[handle + s_dirindex]
  offset += long[handle + s_diroffset]
  if offset > block_size - 4
    ser.dbprintf1(string("WriteDirectory: ERROR - bad offset %d"), offset)
    return
  WriteEEPromLong(index + offset, value)

PRI ReadDirectoryValue(handle, offset) | index
{{
  This routine reads a long value in the directory entry for a
  file.  The byte offset from the beginning of the directory entry is
  indicated by the value of "offset".  This would have a value of
  zero for the file's starting block address and four for the file
  size.
}}
  index := long[handle + s_dirindex]
  offset += long[handle + s_diroffset]
  if offset > block_size - 4
    ser.dbprintf1(string("ReadDirectory: ERROR - bad offset %d"), offset)
    return
  result := ReadEEPromLong(index + offset)

' This routine reads one byte from the EEPROM
PRI ReadEEPromByte(addr)
  return eeprom.ReadByte(eeprom_pin, device_sel, addr)

' This routine writes one byte to the EEPROM
PRI WriteEEPromByte(addr, value)
  eeprom.WriteByte(eeprom_pin, device_sel, addr, value)
  repeat while (eeprom.WriteWait(eeprom_pin, device_sel, addr))

' This routine reads one long from the EEPROM
PRI ReadEEPromLong(addr) | value
  return eeprom.ReadLong(eeprom_pin, device_sel, addr)  

' This routine writes one long to the EEPROM
PRI WriteEEPromLong(addr, value)
  'ser.dbprintf2(string("WriteEEPromLong: $%x $%x"), addr, value)
  eeprom.WriteLong(eeprom_pin, device_sel, addr, value)
  repeat while(eeprom.WriteWait(eeprom_pin, device_sel, addr))

' This routine writes a block of data to the EEPROM
PRI MemCpyEEProm(addr, ptr, len) | len1
  repeat while len > 0
    len1 := 64 - (addr & 63)
    if (len1 > len)
      len1 := len
    eeprom.WritePage(eeprom_pin, device_sel, addr, ptr, len1)
    repeat while(eeprom.WriteWait(eeprom_pin, device_sel, addr))
    len -= len1
    ptr += len1
    addr += len1

PRI EEPromCopyToMem(addr, ptr, len) | len1
  repeat while len > 0
    len1 := 64 - (addr & 63)
    if (len1 > len)
      len1 := len
    eeprom.ReadPage(eeprom_pin, device_sel, addr, ptr, len1)
    len -= len1
    ptr += len1
    addr += len1

' This routine allocates a block of memory
PRI AllocateBlock | index
  ' Find a free block
  index := first_free
  repeat while (index < DISK_END)
    if (ReadEEPromByte(index) & 1)
      first_free := index + block_size
      return index
    index += block_size
  return 0

' This routine searches for a file and returns it starting address
PRI FindFile(fname, dirhand, pfilesize) | index0, filesize, ptr, filename[6]
  'ser.dbprintf3(string("FindFile: %s $%x $%x"), fname, dirhand, pfilesize)
  if dirhand == 0
    return
  'RewindFile(dirhand)
  'InitRootHandle(dirhand)
  'PrintHandle(dirhand)
  seek(dirhand, 0)
  ClearCached(dirhand)
  'PrintHandle(dirhand)
  repeat
    if read(dirhand, @index0, 4) <> 4
      quit
    read(dirhand, @filesize, 4)
    read(dirhand, @filename, fname_size)
    if strcomp(fname, @filename) and (index0 & attrib_mask)
      'long[dirhand + s_offset]   -= dirent_size
      'long[dirhand + s_position] -= dirent_size
      long[pfilesize] := filesize
      'ser.dbprintf3(string("FindFile: %s found. index = $%x, offset = %d"), fname, index, offset)
      return index0
    'else
      'ser.dbprintf3(string("FindFile: %s doesn't match %s, index0 = $%x"), fname, @filename, index0)

' This routine creates a file
PRI CreateFile(fname, modebits) | filesize
  ser.dbprintf2(string("CreateFile: %s %x"), fname, modebits)
  if FindFile(fname, @currentwd_block, @filesize)
    return
  modebits := (%10000 | modebits) & attrib_mask
  UpdateDirectoryEntry(@currentwd_block, 0, modebits, 0, fname)
  return 1

PRI OpenRead(fname, handle) | index0, offset0, index1, filesize1
  ifnot (index1 := FindFile(fname, @currentwd_block, @filesize1))
    return
  index0 := long[@currentwd_block + s_index]
  offset0 := long[@currentwd_block + s_offset] - dirent_size
  InitFileHandle(handle, index0, offset0, index1, filesize1, handle + handle_size)
  return 1

PRI OpenAppend(fname, handle) | index0, offset0, index1, filesize1, next1
  if (index1 := FindFile(fname, @currentwd_block, @filesize1))
    index0 := long[@currentwd_block + s_index]
    offset0 := long[@currentwd_block + s_offset] - dirent_size
    InitFileHandle(handle, index0, offset0, index1, filesize1, handle + handle_size)
    seek(handle, filesize1)
    'repeat while read(handle, @currentwd_block, handle_size) == handle_size
  else
    UpdateDirectoryEntry(@currentwd_block, handle, %10110, 0, fname)
  return 1

PRI OpenWrite(fname, handle) | index0, index1, filesize1
  if (index1 := FindFile(fname, @currentwd_block, @filesize1))
  ' Free up all the blocks
    index1 &= index_mask
    repeat while (index1)
      index0 := ReadEEPromLong(index1)
      WriteEEPromLong(index1, 1)
      if index1 < first_free
        first_free := index1
      index1 := index0
    long[@currentwd_block + s_position] -= dirent_size
    long[@currentwd_block + s_offset] -= dirent_size
    UpdateDirectoryEntry(@currentwd_block, handle, %10110, 0, 0)
  else
    UpdateDirectoryEntry(@currentwd_block, handle, %10110, 0, fname)
  return 1

PRI UpdateDirectoryEntry(dirhand, handle, index, filesize, fname) | dirindex, diroffset, written
  write(dirhand, @index, 4)
  write(dirhand, @filesize, 4)
  if fname
    write(dirhand, fname, fname_size)
    written := dirent_size
  else
    written := 8
  flush(dirhand)
  if handle
    dirindex := long[dirhand + s_index]
    diroffset := long[dirhand + s_offset] - written
    InitFileHandle(handle, dirindex, diroffset, index, filesize, handle + handle_size) ' Should specify buffer