  'XMODEM chars from ymodem.txt
  SOH=$01
  STX=$02
  EOT=$04
  ACK=$06
  NAK=$15
  CAN=$18
  Cee=$43  'liberties here
  Zee=$1A  'or SUB  (DOS EOF?)

  MSG_TIMEOUT  = -2
  MSG_EOT      = -3
  BYTE_TIMEOUT = -4
  MSG_BADNUM   = -5
  MSG_CRCERR   = -6

OBJ  'Objects Section
  ser : "cserial"
  c : "clib"

'var
'   byte byte[pdata]1028] 'packet data
dat
  pdata long 0

PUB SendFile(argc, argv) | infile, filesize, val, len, fname, packetnum
  pdata := c.malloc(1028)
  ifnot pdata
    c.printf0(string("Could not malloc 1028 bytes for the data buffer\n"))
    return
  fname := long[argv][1]
  infile := c.fopen(fname, string("r"))
  ifnot infile
    c.printf1(string("Could not open %s\n"), fname)
    c.free(pdata)
    return
  filesize := c.getfilesize(infile)
  c.memset(pdata, 0, 130)
  byte[pdata][1] := 255
  c.sprintf2(pdata+2, string("%s %d"), fname, filesize)
  len := strsize(fname)
  byte[pdata][len+2] := 0
  repeat
    val := ser.rxtime(30000)
    if val == -1 or val == 27
      c.fclose(infile)
      c.free(pdata)
      return
    if val == CEE
      quit

  SendPacket(0, pdata, 128)
      
  repeat
    val := ser.rxtime(10000)
    if val == -1 or val == 27
      c.fclose(infile)
      c.free(pdata)
      return
    if val == CEE
      quit

  packetnum := 1
  repeat while filesize > 0
    len := filesize
    if len > 1024
      len := 1024
    filesize -= len
    c.fread(pdata + 2, 1, len, infile)
    SendPacket(packetnum, pdata, len)
    packetnum++

  SendEOT
  byte[pdata][2] := 0
  SendPacket(0, pdata, 128)

  c.fclose(infile)
  c.free(pdata)

PUB SendEOT | val
  repeat
    ser.tx(EOT)    
    repeat
      val := ser.rxtime(10000)
      if val == ACK
        return 1
      elseif val == NAK
        quit

PUB SendPacket(packetnum, ptr, num) | i, val, firstchar
  if (num =< 0) or (num > 1024)
    return
  elseif (num =< 128)
    num := 128
    firstchar := SOH
  else
    num := 1024
    firstchar := STX
  byte[ptr] := packetnum
  byte[ptr][1] := 255 - packetnum
  AppendCRC(ptr+2, num)
  repeat
    ser.tx(firstchar)
    repeat i from 0 to num + 3
      ser.tx(byte[ptr][i])
    repeat
      val := ser.rxtime(10000)
      if val == ACK
        return num
      elseif val == NAK
        quit
  return num

PUB ReceiveFile | packetnum, num, ptr, filesize, fname[10], outfile
  pdata := c.malloc(1028)
  ifnot pdata
    c.printf0(string("Could not malloc 1028 bytes for the data buffer\n"))
    return
  c.printf0(string("ymodem receive starts in 10 seconds\n"))
  waitcnt(clkfreq*10+cnt)
  outfile := 0
  ptr := pdata
  packetnum := 0
  repeat
    if packetnum < 2
      ser.tx(CEE)
    num := ReceivePacket(ptr)
    'c.fprintf2(logfile, string("ReceivePacket returned %d %d\n"), num, byte[pdata][0])
    if num == MSG_TIMEOUT
      ser.tx(EOT)
      quit
    elseif num == MSG_EOT
      packetnum := 0
      ser.tx(ACK)
      next
    elseif num =< 0
      ser.tx(NAK)
      next
    if packetnum == 0
      if byte[pdata][2] == 0
        ser.tx(ACK)
        quit
      filesize := DecodeFirstPacket(ptr+2, num, @fname)
      'c.fprintf2(logfile, string("fname = %s, size = %d\n"), @fname, filesize)
      outfile := c.fopen(@fname, string("w"))
      'Open file
    else
      if num > filesize
        num := filesize
      c.fwrite(@byte[pdata][2], 1, num, outfile)
      filesize -= num
    ser.tx(ACK)
    packetnum++
  if outfile
    c.fclose(outfile)
  c.free(pdata)

PUB DecodeFirstPacket(ptr, num, fname) | i, val, filesize
  ifnot byte[ptr]
    byte[fname] := 0
    return -1
  repeat i from 0 to 40
    byte[fname][i] := byte[ptr][i]
    ifnot byte[ptr][i]
      i++
      quit
  filesize := 0
  repeat while i < num
    val := byte[ptr][i++]
    if val < "0" or val > "9"
      quit
    filesize := (filesize * 10) + val - "0"

  return filesize
                 
PUB ReceivePacket(ptr) | num, val, timeout, ptr0
  ptr0 := ptr
  timeout := (clkfreq * 10) + cnt
  ' Get start of packet
  repeat
    repeat while (val := ser.rxcheck) == -1
      if cnt - timeout => 0
        return MSG_TIMEOUT
    'c.fprintf1(logfile, string("rx %02x\n"), val)
    if val == SOH
      num := 128
      quit
    elseif val == STX
      num := 1024
      quit
    elseif val == EOT
      return MSG_EOT
      
  ' Read the packet    
  repeat num + 4
    timeout := clkfreq + cnt
    repeat while (val := ser.rxcheck) == -1
      if cnt - timeout => 0
        return BYTE_TIMEOUT
    byte[ptr++] := val
    
  ' Check for valid packet number pair
  if (byte[ptr0] + byte[ptr0][1]) <> 255
    return MSG_BADNUM
    
  ' Check the CRC
  ifnot CheckCRC(ptr0+2, num)
    return MSG_CRCERR
    
  return num

PUB CheckCRC(ptr, num) | crc
  crc := ComputeCRC(ptr, num)
  return (byte[ptr][num] == (crc >> 8)) and (byte[ptr][num+1] == (crc & $ff))

PUB AppendCRC(ptr, num) | crc
  crc := ComputeCRC(ptr, num)
  byte[ptr][num] := crc >> 8
  byte[ptr][num+1] := crc & $ff

PUB ComputeCRC(ptr, num)
  repeat num
    result ^= (byte[ptr++] << 8)
    repeat 8
      result <<= 1
      if result & $10000
        result ^= $1021
  result &= $ffff